﻿--[[
    Author: Valana <Team BATTLE> of Mug'thol Alliance-US
]]

local TbChatEnhancements = LibStub("AceAddon-3.0"):NewAddon("TbChatEnhancements", "AceEvent-3.0", "AceHook-3.0", "LibWho-2.0")
local L = LibStub("AceLocale-3.0"):GetLocale("TB_ChatEnhancements", true)
local ColorPickerFrame = ColorPickerFrame
local TbCore = TbCore
local WhoFrame = WhoFrame
local WhoFrameEditBox = WhoFrameEditBox

local _G = _G
local ERR_GUILD_PERMISSIONS = ERR_GUILD_PERMISSIONS
local strfind = strfind
local strsub = strsub
local gsub = gsub
local strmatch = strmatch
local format = format
local strupper = strupper
local tinsert = tinsert
local ipairs = ipairs
local date = date
local time = time
local wipe = wipe
local IsShiftKeyDown = IsShiftKeyDown
local IsControlKeyDown = IsControlKeyDown
local UnitIsUnit = UnitIsUnit
local UnitName = UnitName
local UnitClass = UnitClass
local GetNumWhoResults = GetNumWhoResults
local GetNumFriends = GetNumFriends

local _player = UnitName("player")
local _isSendingChatMessage
local _copyChatItemReplacement
local _copyChatAchievementReplacement
local _copyChatSpellReplacement
local _findCopyChatPrefix
local _findSpellPrefix
local _findAchievementPrefix
local _findItemPrefix
local _afkBuffer = { }
local _highlightPatterns
local _highlightHex = ""
local _chatTimeStampsFormat = ""
local _classColorsHex = { }
local _isProcessingGuildRosterUpdate
local _whoData = { ["num"] = 0, ["text"] = "", ["time"] = 0 }
local _numGroupMembers = 0
local _numTeamMates = 0
local _numGuildMates = 0
local _numFriends = 0
local _iconMatcher =
{
    ["RaidTargetingIcon_1"] = "{rt1}",
    ["RaidTargetingIcon_2"] = "{rt2}",
    ["RaidTargetingIcon_3"] = "{rt3}",
    ["RaidTargetingIcon_4"] = "{rt4}",
    ["RaidTargetingIcon_5"] = "{rt5}",
    ["RaidTargetingIcon_6"] = "{rt6}",
    ["RaidTargetingIcon_7"] = "{rt7}",
    ["RaidTargetingIcon_8"] = "{rt8}",
    ["ChatIcon-Blizz"] = "<GM>",
}

local _lastWho = 0
local _whoOptions =
{
    queue = TbChatEnhancements.WHOLIB_QUEUE_QUIET,
    timeout = 0,
    callback = "OnWhoLibWho",
    handler = TbChatEnhancements,
    flags = 0,
}
local _userInfoOptions =
{
    queue = TbChatEnhancements.WHOLIB_QUEUE_QUIET,
    timeout = 0,
    callback = "OnWhoLibUserInfo",
    handler = TbChatEnhancements,
    flags = 0,
}


-------------------------------------------------------------------------------
-- Add time stamps to chat messages.
-------------------------------------------------------------------------------
function TbChatEnhancements:AddMessage(frame, message, ...)
    if (not message or not strfind(message, "[^%s]")) then return end

    --
    -- Ignore "No player named 'X' is currently playing" and "You don't have permission to do that" messages.
    --
    if (strfind(message, L["TARGET_NOT_FOUND"]) and (self:IsAutoBGWhispering() or self:IsCrossGuildChatWhispering())) then return end
    if (message == ERR_GUILD_PERMISSIONS and self:IsCrossGuildChatWhispering()) then return end

    --
    -- First save off our original message string before we add even more junk to it.
    --
    local linkBody = message

    --
    -- Highlight the names of players we know the classes of. Note: do this before custom highlighting to give it precedence.
    --
    if (self:IsPerClassHighlighting()) then
        message = gsub(message, "^%f[%a\128-\255]([%a\128-\255]+)%f[^%a\128-\255]", function(player) return self:GetPerClassHighlightedName("", player) end)                -- Valana has gone offline.
        message = gsub(message, "(%d?%d%p+%s+)([%a\128-\255]+)([%p%s]+)", function(lead, player, trail) return self:GetPerClassHighlightedName(lead, player, trail) end)	-- Recount summaries.
        message = gsub(message, "\124h(.-)\124h", function(player) return self:GetPerClassHighlightedName("\124h", player, "\124h") end)                                    -- Valana performs a custom emote.
        message = gsub(message, "%[(.-)%]", function(player) return self:GetPerClassHighlightedName("[", player, "]") end)                                                  -- [Valana] has come online.
        message = gsub(message, L["ROLL_BY"], function(lead, player) return self:GetPerClassHighlightedName(lead, player) end)                                              -- Need Roll - 100 for [Envoy of Mortality] by Valana.
        message = gsub(message, L["DUEL_WIN_PARTIAL"], function(lead, player, trail) return self:GetPerClassHighlightedName(lead, player, trail) end)                       -- Litto has defeated Valana in a duel.
        message = gsub(message, L["DUEL_FLED_PARTIAL"], function(lead, player, trail) return self:GetPerClassHighlightedName(lead, player, trail) end)                      -- Valana has fled from Valana in a duel.
    end

    --
    -- Apply custom coloring.
    --
    for index, word in ipairs(self.db.profile.highlightWords) do
        local pattern = _highlightPatterns[TbCore:GetCaseInsensitivePattern(word)]
        message = gsub(message, "\124c%x%x%x%x%x%x%x%x(" .. pattern .. ")\124r", format(_highlightHex, "", "%1", ""))				-- Strip any previous coloring.
        message = gsub(message, "%f[%a1-9\128-\255:](" .. pattern .. ")%f[^%a1-9\128-\255]", format(_highlightHex, "", "%1", ""))	-- Valana has gone offline. / [Valana] has come online.
        message = gsub(message, "\124h(" .. pattern .. ")\124h", format(_highlightHex, "\124h", "%1", "\124h"))						-- Valana performs a custom emote.
    end

    --
    -- Build the chat time stamp hyperlink body, stripping all escape codes and replacing icon lookup codes with their chat equivalents.
    --
    if (self:GetChatTimeStampsFormat() > 1) then
        --
        -- Do not gsub something like "|c%x%x%x%x%x%x%x%x(.-)|r" because there may be color-coded text embedded inside more color-coded text.
        --
        linkBody = gsub(linkBody, "\124c%x%x%x%x%x%x%x%x", "")
        linkBody = gsub(linkBody, "\124r", "")
        linkBody = gsub(linkBody, "\124Hitem:(%d-):.-\124h.-\124h", _copyChatItemReplacement)
        linkBody = gsub(linkBody, "\124Hachievement:(%d-):.-\124h.-\124h", _copyChatAchievementReplacement)
        linkBody = gsub(linkBody, "\124Hspell:(%d-)\124h(.-)\124h", _copyChatSpellReplacement)
        linkBody = gsub(linkBody, "\124H.-\124h(.-)\124h", "%1")
        linkBody = gsub(linkBody, "\124T.-%-(.-)%.blp.-\124t", _iconMatcher)
        linkBody = gsub(linkBody, "\124", "/")

        --
        -- Add the hyperlink to the chat time stamp, which will contain the unformatted text of the timestamped message.
        --
        local t = date(_chatTimeStampsFormat)
        message = format("\124H%s%s %s\124h%s\124h %s", self:GetCopyChatPrefix(), t, linkBody, t, message)
    end

    --
    -- Forward our modified message to our original handler.
    --
    return self.hooks[frame].AddMessage(frame, message, ...)
end

function TbChatEnhancements:GetPerClassHighlightedName(lead, player, trail)
    local data = self.db.realm.players[player] or self.db.realm.players[TbCore:Capitalize(player)]
    if (data and data["class"]) then	-- The data["class"] needs to be here due to corrupt databases I believe...
		data["time"] = time()
        return format(_classColorsHex[data["class"]], lead or "", player, trail or "")
    end
end


-------------------------------------------------------------------------------
-- Toggle our AFK and DND flags while talking.
-------------------------------------------------------------------------------
function TbChatEnhancements:SendChatMessage(message, distribution, ...)
    --
    -- Figure out whether we are marked AFK or DND, unless we are trying to get out of it.
    --
    distribution = strupper(distribution)
    local status
    if (distribution ~= "AFK" and distribution ~= "DND") then
        if (UnitIsAFK("player")) then status = "AFK"
        elseif (UnitIsDND("player")) then status = "DND" end
    end

    --
    -- Mark that we are sending an emote to filter out the system announcement, and clear our AFK/DND status.
    --
    if (status) then
        _isSendingChatMessage = true
        self.hooks["SendChatMessage"]("", status, ...)
    end

    --
    -- Send out our chat message.
    --
    local rtn = self.hooks["SendChatMessage"](message, distribution, ...)

    --
    -- Restore our AFK/DND status if needed.
    --
    if (status == "DND" or (status == "AFK" and tonumber(GetCVar("autoClearAFK")) == 0)) then
        self.hooks["SendChatMessage"](self.db.char.afkMessage, status, ...)
        _isSendingChatMessage = nil
    end

    --
    -- Return whatever the regular call would have returned.
    --
    return rtn
end


-------------------------------------------------------------------------------
-- Chat frame hooks.
-------------------------------------------------------------------------------
--
-- Enables mouse scrolling for chat frames.
--
function TbChatEnhancements:OnMouseWheel(frame, direction, ...)
    if (direction > 0) then
        if (IsShiftKeyDown()) then
            frame:ScrollToTop()
        elseif (IsControlKeyDown()) then
            frame:PageUp()
        else
            frame:ScrollUp()
        end
    else
        if (IsShiftKeyDown()) then
            frame:ScrollToBottom()
        elseif (IsControlKeyDown()) then
            frame:PageDown()
        else
            frame:ScrollDown()
        end
    end
end

--
-- Intercept hyperlink clicks.
--
function TbChatEnhancements:OnHyperlinkClick(frame, link, ...)
    --
    -- Perform some special handling if we find the TbCCP: prefix.
    --
    local stripped = strmatch(link, _findCopyChatPrefix)
    if (not stripped) then return self.hooks[frame].OnHyperlinkClick(frame, link, ...) end

    if (IsShiftKeyDown() and ChatFrameEditBox:IsVisible()) then
        --
        -- Replace spells with their actual links.
        --
        stripped = gsub(stripped, _findSpellPrefix,
            function(id, name)
                if (id) then
                    return GetSpellLink(id) or name
                end
            end)

        --
        -- Replace achievements with their actual links.
        --
        stripped = gsub(stripped, _findAchievementPrefix, function(id) if (id) then return GetAchievementLink(id) end end)

        --
        -- Replace items with their actual links.
        --
        stripped = gsub(stripped, _findItemPrefix,
            function(id)
                if (id) then
                    local name, link = GetItemInfo(id)
                    if (link) then return link end
                    return name
                end
            end)
        ChatFrameEditBox:Insert(stripped)
    end
end


-------------------------------------------------------------------------------
-- Suppress annoying miscellaneous messages.
-------------------------------------------------------------------------------
--
-- Do not show outgoing messages to oneself.
--
TbChatEnhancements.WhisperInformFilter = function(frame, e, message, sender, ...)
    if (UnitIsUnit("player", sender)) then return true end
end

--
-- Ignore repeated AFK and DND messages.
--
TbChatEnhancements.AfkMessageFilter = function(frame, e, message, sender, ...)
    --
    -- Don't display it if we're talking to ourselves. Silent Ignore prevents us from sending a message to someone we are silently-
    -- ignoring, and we don't get responses from players blocked through the Blizzard interface.
    --
    if (UnitIsUnit("player", sender)) then return true end

    --
    -- Initialize an entry for this player in our buffer if needed.
    --
    if (not _afkBuffer[sender]) then _afkBuffer[sender] = { ["time"] = 0, ["message"] = "" } end

    --
    -- Determine whether we should display this response. We ignore identical, consecutive AFK and DND messages sent within 5 minutes
    -- of the previous. This means if someone goes from "AFK: Away from Keyboard" to "DND: Away from Keyboard", we will ignore the
    -- DND messages also (in other words, we treat AFK and DND messages the same).
    --
    local t = time()
    local rtn = t - _afkBuffer[sender]["time"] < 300 and _afkBuffer[sender]["message"] == message

    --
    -- Update our data regarding this player.
    --
    _afkBuffer[sender]["time"], _afkBuffer[sender]["message"] = t, message
    return rtn
end

--
-- Ignore local achievement announcements from people who are not on our friends list.
--
TbChatEnhancements.AchievementMessageFilter = function(frame, e, message, sender, ...)
    --
    -- Only show achievement messages from ourselves and our friends if they are not also being ignored. Also filter out achievement
    -- announcements from guildmates because we will get those again through the guild channel.
    --
    if (TbCore:IsIgnored(sender) or TbCore:IsGuildMate(sender) or (not UnitIsUnit("player", sender) and not TbCore:IsFriend(sender))) then return true end
end

--
-- Ignore guild achievement announcements from people we are ignoring.
--
TbChatEnhancements.GuildAchievementMessageFilter = function(frame, e, message, sender, ...)
    if (TbCore:IsIgnored(sender)) then return true end
end

TbChatEnhancements.SystemMessageFilter = function(frame, e, message, ...)
    --
    -- Filter out AFK/DND announcement messages if we're sending a chat message.
    --
    local match = strmatch(message, L["GOING_AFK_OR_DND"])
    if (match) then
        --
        -- Remember our AFK or DND message.
        --
        TbChatEnhancements.db.char.afkMessage = match
        return _isSendingChatMessage
    elseif (message == CLEARED_AFK or message == CLEARED_DND) then return _isSendingChatMessage end

    --
    -- Filter out annoying messages.
    --
    if (message == ERR_RAID_YOU_LEFT) then return true end
    if (strfind(message, L["TIPSY"])) then return true end
    if (strfind(message, L["TIPSY_ITEM"])) then return true end
    if (strfind(message, L["DRUNK"])) then return true end
    if (strfind(message, L["DRUNK_ITEM"])) then return true end
    if (strfind(message, L["SOBER"])) then return true end
    if (strfind(message, L["SOBER_ITEM"])) then return true end
    if (strfind(message, L["SMASHED"])) then return true end
    if (strfind(message, L["SMASHED_ITEM"])) then return true end

    --
    -- Only filter out duel outcomes if they did not involve us.
    --
    local str1, str2 = strmatch(message, L["DUEL_WIN"])
    if (str1 and not (UnitIsUnit(_player, str1) or UnitIsUnit("target", str1) or TbCore:IsFriendOrGuildMate(str1) or UnitIsUnit(_player, str2) or UnitIsUnit("target", str2) or TbCore:IsFriendOrGuildMate(str2))) then return true end
    str1, str2 = strmatch(message, L["DUEL_FLED"])
    if (str1 and not (UnitIsUnit(_player, str1) or UnitIsUnit("target", str1) or TbCore:IsFriendOrGuildMate(str1) or UnitIsUnit(_player, str2) or UnitIsUnit("target", str2) or TbCore:IsFriendOrGuildMate(str2))) then return true end
end

TbChatEnhancements.ChannelMessageFilter = function(frame, e, message, ...)
    if (message == "YOU_CHANGED") then return true end
    if (message == "SUSPENDED") then return true end
    if (message == "SET_MODERATOR") then return true end
    if (message == "OWNER_CHANGED") then return true end
    if (message == "YOU_JOINED") then return true end
    if (message == "YOU_LEFT") then return true end
end

TbChatEnhancements.ChannelJoinFilter = function(frame, e, ...)
    return true
end


-------------------------------------------------------------------------------
-- Update our player data every time new data becomes available.
-------------------------------------------------------------------------------
function TbChatEnhancements:SavePlayerData(name, class, realm)
    if (name and class and not (class == "" or class == UNKNOWN)) then
        if (realm and realm ~= "") then name = name .. "-" .. realm end
        if (not self.db.realm.players[name]) then
            self.db.realm.players[name] = { }
        end
        self.db.realm.players[name]["class"] = class
        self.db.realm.players[name]["time"] = time()
        return true
    end
end

function TbChatEnhancements:CHAT_MSG_SYSTEM(e)
    --
    -- If appropriate and the Who frame is visible, we got a WHO_LIST_UPDATE event so just ignore this.
    --
    if (WhoFrame:IsVisible()) then return end
    local num, text, t = GetNumWhoResults(), WhoFrameEditBox:GetText(), time()
    if (num ~= _whoData["num"] or text ~= _whoData["text"] or t - _whoData["time"] > 60) then
        --
        -- Only process this system event if the number of results or the query is different from last time.
        -- We may miss some valid updates, but it is better than re-processing the entire who result set on
        -- every single CHAT_MSG_SYSTEM event. We will still process it after a minute if it's still there.
        --
        _whoData["num"], _whoData["text"], _whoData["time"] = num, text, t
        self:WHO_LIST_UPDATE()
    end
end

function TbChatEnhancements:UPDATE_MOUSEOVER_UNIT(e)
    if (UnitIsPlayer("mouseover")) then
        local name, realm = UnitName("mouseover")
        local _, class = UnitClass("mouseover")
        self:SavePlayerData(name, class, realm)
    end
end

function TbChatEnhancements:PLAYER_TARGET_CHANGED(e)
    if (UnitIsPlayer("target")) then
        local name, realm = UnitName("target")
        local _, class = UnitClass("target")
        self:SavePlayerData(name, class, realm)
    end
end

function TbChatEnhancements:WHO_LIST_UPDATE(e)
    local numWho = GetNumWhoResults()
    for i = 1, numWho do
        local name, _, _, _, _, _, class = GetWhoInfo(i)
        self:SavePlayerData(name, class)
    end
end

function TbChatEnhancements:PARTY_MEMBERS_CHANGED(e)
    --
    -- Refresh our data if we have more group members than during the previous check.
    --
    if (TbCore:IsInBattleground()) then return end

    local numGroupMembers
    if (TbCore:IsInRaid()) then
        numGroupMembers = GetNumRaidMembers() - 1
        if (numGroupMembers > _numGroupMembers) then
            for i = 1, numGroupMembers do
                --
                -- If somehow we failed to save the player's data (eg. if some raid member's class comes back as unknown), just
                -- abort so we may re-parse the entire roster next time.
                --
                local name, _, _, _, _, class = GetRaidRosterInfo(i)
                if (not self:SavePlayerData(name, class)) then return end
            end
        end
    else
        numGroupMembers = GetNumPartyMembers()
        if (numGroupMembers > _numGroupMembers) then
            for i = 1, numGroupMembers do
                local current = "party" .. i
                local name, realm = UnitName(current)
                local _, class  = UnitClass(current)
                if (not self:SavePlayerData(name, class, realm)) then return end
            end
        end
    end
    _numGroupMembers = numGroupMembers
end

function TbChatEnhancements:UPDATE_BATTLEFIELD_STATUS(e)
    --
    -- Refresh our data if we have more battleground teammates online than during the previous check.
    --
    local numTeamMates = 0
    if (TbCore:IsInBattleground()) then
        numTeamMates = GetNumBattlefieldScores()
        if (numTeamMates > _numTeamMates) then
            for i = 1, numTeamMates do
                local name, _, _, _, _, _, _, _, _, class = GetBattlefieldScore(i)
                if (not self:SavePlayerData(name, class)) then return end
            end
        end
    end
    _numTeamMates = numTeamMates
end

function TbChatEnhancements:GUILD_ROSTER_UPDATE(e, isLocal)
    --
    -- When a player joins or leaves the guild, 1 is passed as the value of isLocal and we do not automatically get the updated
    -- guild roster. TB Core will take care of requesting the guild roster.
    --
    if (isLocal) then return end

    --
    -- Changing whether we are showing offline members will cause this event to re-enter, so set a flag.
    --
    if (_isProcessingGuildRosterUpdate) then return end
    _isProcessingGuildRosterUpdate = true
    local numGuildMates = GetNumGuildMembers(true)
    if (numGuildMates > _numGuildMates) then
        --
        -- Refresh our data if we have more guildmates online than during the previous check.
        --
        local wasShowingOffling = GetGuildRosterShowOffline()
        SetGuildRosterShowOffline(true)
        for i = 1, numGuildMates do
            local name, _, _, _, _, _, _, _, _, _, class = GetGuildRosterInfo(i)
            if (not self:SavePlayerData(name, class)) then
                --
                -- If saving the guild mate's data failed, exit without updating our guild mate count. This was most likely because roster
                -- information was not yet fully available.
                --
                SetGuildRosterShowOffline(wasShowingOffline)
                _isProcessingGuildRosterUpdate = nil
                return
            end
        end
        SetGuildRosterShowOffline(wasShowingOffline)
    end
    _numGuildMates = numGuildMates
    _isProcessingGuildRosterUpdate = nil
end

function TbChatEnhancements:FRIENDLIST_UPDATE(e)
    --
    -- Refresh our data if we have more friends than during the previous check.
    --
    local numFriends = GetNumFriends()
    if (numFriends > _numFriends) then
        for i = 1, numFriends do
            local name, _, class = GetFriendInfo(i)
            if (not self:SavePlayerData(name, TbCore:GetUnlocalizedClass(class))) then return end
        end
    end
    _numFriends = numFriends
end


-------------------------------------------------------------------------------
-- Automatically perform a /who query against players whose classes are unknown.
-------------------------------------------------------------------------------
function TbChatEnhancements:ZONE_CHANGED_NEW_AREA(e)
    if (TbCore:IsInBattleground()) then return end

    local t = time()
    if (t - _lastWho > 10) then
        _lastWho = t
        self:Who(GetRealZoneText(), _whoOptions)
    end
end

function TbChatEnhancements:OnWhoLibWho(query, results, ...)
    for index, result in ipairs(results) do
        self:SavePlayerData(result.Name, result.NoLocaleClass)
    end
end

function TbChatEnhancements:CHAT_MSG_CHANNEL(e, message, sender, language, arg4, arg5, flag, ...)
    --
    -- Only send a /who query against players who they talk on open channels or whisper, to guarantee that they are not from the opposite
    -- faction. Limit this to a rate of 1 /who query every 10 seconds at most.
    --
    if (flags == "GM") then return end

    local t = time()
    if (sender and sender ~= "" and not self.db.realm.players[sender] and t - _lastWho > 10) then
        _lastWho = t
        self:UserInfo(sender, _userInfoOptions)
    end
end

function TbChatEnhancements:OnWhoLibUserInfo(result, ...)
	self:SavePlayerData(result.Name, result.NoLocaleClass)
end


-------------------------------------------------------------------------------
-- Perform initialization.
-------------------------------------------------------------------------------
function TbChatEnhancements:OnInitialize()
    --
    -- Initialize some member variables.
    --
    _copyChatItemReplacement = self:GetCopyChatItemPrefix() .. "%1:"
    _copyChatAchievementReplacement = self:GetCopyChatAchievementPrefix() .. "%1:"
    _copyChatSpellReplacement = self:GetCopyChatSpellPrefix() .. "%1:%2:"
    _findCopyChatPrefix = self:GetCopyChatPrefix() .. "(.*)"
    _findSpellPrefix = self:GetCopyChatSpellPrefix() .. "(%d+):(.-):"
    _findAchievementPrefix = self:GetCopyChatAchievementPrefix() .. "(%d+):"
    _findItemPrefix = self:GetCopyChatItemPrefix() .. "(%d+):"
    for class, color in pairs(CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS) do
        _classColorsHex[class] = format("%%s\124c%s%%s\124r%%s", TbCore:RgbToHex(color.r, color.g, color.b))
    end

    --
    -- This metatable stores the patterns used to match the text to be highlighted.
    --
    _highlightPatterns = setmetatable({ }, { __index =
        function(tbl, word)
            local pattern = gsub(word, "%*", "%[%%a\\128-\\255%]*")
            rawset(tbl, word, pattern)
            return pattern
        end })

    --
    -- Watch for when the GMChatUI gets loaded so we may adjust its chat frame.
    --
    self:RegisterEvent("ADDON_LOADED")

    --
    -- Load our settings.
    --
    local r, g, b = TbCore:ColorToRgb("LTBLUE")
    self.db = LibStub("AceDB-3.0"):New("TbChatEnhancementsDb",
    {
        global =
        {
            chatEditBoxLocation = 2,
            chatTimeStampsFormat = 20,
            extendChatHistory = true,
            fadeChat = false,
            filterChannelAnnouncements = true,
            highlightColor = { ["r"] = r, ["g"] = g, ["b"] = b },
            mouseScroll = true,
            perClassHighlighting = true,
            showChatFrameButtons = true,
        },
        realm = { players = { } },
        profile = { highlightWords = { } },
        char = { afkMessage = "" },
    }, "Default")
    self.db:SetProfile("Default")

    --
    -- Remove outdated entries.
    --
    self:CleanUpPlayerCache()

    --
    -- Add the player to the database. We do it every time instead of just setting a default, in case the player gets a name change.
    --
    local _, class = UnitClass("player")
    self:SavePlayerData(_player, class)

    --
    -- Create our options menus.
    --
    local _, addon = GetAddOnInfo("TB_ChatEnhancements")
    LibStub("AceConfig-3.0"):RegisterOptionsTable(addon,
        {
            type = "group",
            name = addon,
            desc = GetAddOnMetadata("TB_ChatEnhancements", L["DESCRIPTION_METADATA_FIELD"]),
            handler = TbChatEnhancements,
            childGroups  = "tab",
            args =
            {
                subHeader =
                {
                    order = 0,
                    type = "description",
                    name = L["MENU_SUBHEADER"],
                },
                chatFrame =
                {
                    order = 1,
                    type = "group",
                    name = L["MENU_CHAT_FRAME"],
                    desc = L["MENU_CHAT_FRAME_DESCRIPTION"],
                    args =
                    {
                        chatFrameSubHeader =
                        {
                            order = 0,
                            type = "description",
                            name = L["MENU_CHAT_FRAME_SUBHEADER"],
                        },
                        extendChatHistory =
                        {
                            order = 10,
                            type = "toggle",
                            name = L["MENU_EXTEND_HISTORY"],
                            desc = L["MENU_EXTEND_HISTORY_DESCRIPTION"],
                            get = "IsExtendingChatHistory",
                            set = "ToggleExtendChatHistory",
                        },
                        fadeChat =
                        {
                            order = 20,
                            type = "toggle",
                            name = L["MENU_FADE_CHAT"],
                            desc = L["MENU_FADE_CHAT_DESCRIPTION"],
                            get = "IsFadingChat",
                            set = "ToggleFadeChat",
                        },
                        showChatFrameButtons =
                        {
                            order = 30,
                            type = "toggle",
                            name = L["MENU_SHOW_CHAT_FRAME_BUTTONS"],
                            desc = L["MENU_SHOW_CHAT_FRAME_BUTTONS_DESCRIPTION"],
                            get = "IsShowingChatFrameButtons",
                            set = "ToggleShowChatFrameButtons",
                        },
                        mouseScroll =
                        {
                            order = 40,
                            type = "toggle",
                            name = L["MENU_MOUSE_SCROLL"],
                            desc = L["MENU_MOUSE_SCROLL_DESCRIPTION"],
                            get = "IsMouseScrolling",
                            set = "ToggleMouseScroll",
                        },
                        filterChannelAnnouncements =
                        {
                            order = 50,
                            type = "toggle",
                            name = L["MENU_FILTER_CHANNEL_ANNOUNCEMENTS"],
                            desc = L["MENU_FILTER_CHANNEL_ANNOUNCEMENTS_DESCRIPTION"],
                            get = "IsFilteringChannelAnnouncements",
                            set = "ToggleFilterChannelAnnouncements",
                        },
                        dropdownsSeparator =
                        {
                            order = 59,
                            type = "description",
                            name = "",
                        },
                        showChatTimeStamps =
                        {
                            order = 60,
                            type = "select",
                            name = L["MENU_SHOW_CHAT_TIMESTAMPS"],
                            desc = L["MENU_SHOW_CHAT_TIMESTAMPS_DESCRIPTION"],
                            style = "dropdown",
                            values =
                            {
                                [1] = L["NONE"],
                                [5] = L["[12S]"],
                                [10] = L["12S>"],
                                [15] = L["[24S]"],
                                [20] = L["24S>"],
                            },
                            get = "GetChatTimeStampsFormat",
                            set = "SetChatTimeStampsFormat",
                        },
                        chatEditBoxLocation =
                        {
                            order = 61,
                            type = "select",
                            name = L["MENU_CHAT_EDIT_BOX_LOCATION"],
                            desc = L["MENU_CHAT_EDIT_BOX_LOCATION_DESCRIPTION"],
                            style = "dropdown",
                            values =
                            {
                                [1] = L["TOP"],
                                [2] = L["BOTTOM"],
                            },
                            get = "GetChatEditBoxLocation",
                            set = "SetChatEditBoxLocation",
                        },
                    },
                },
                textHighlighting =
                {
                    order = 100,
                    type = "group",
                    name = L["MENU_TEXT_HIGHLIGHTING"],
                    desc = L["MENU_TEXT_HIGHLIGHTING_DESCRIPTION"],
                    args =
                    {
                        textHighlightingSubHeader =
                        {
                            order = 0,
                            type = "description",
                            name = L["MENU_TEXT_HIGHLIGHTING_SUBHEADER"]
                        },
                        wordList =
                        {
                            order = 10,
                            type = "input",
                            name = L["MENU_LIST_TEXT_HIGHLIGHTING"],
                            desc = L["MENU_LIST_TEXT_HIGHLIGHTING_DESCRIPTION"],
                            usage = L["MENU_LIST_TEXT_HIGHLIGHTING_USAGE"],
                            width = "full",
                            multiline = true,
                            -- pattern = "[%w+[,;%s]]+",
                            get = "GetHighlightWords",
                            set = "SetHighlightWords",
                        },
                        pickColor =
                        {
                            order = 11,
                            type = "execute",
                            name = L["MENU_PICK_COLOR"],
                            desc = L["MENU_PICK_COLOR_DESCRIPTION"],
                            func = "PickHighlightColor",
                        },
                        test =
                        {
                            order = 12,
                            type = "execute",
                            name = L["MENU_TEST_HIGHLIGHTING"],
                            desc = L["MENU_TEST_HIGHLIGHTING_DESCRIPTION"],
                            func = function() TbCore:Print(gsub(self:GetHighlightWords(), "%*", "z")) end,
                        },
                        perClassHighlighting =
                        {
                            order = 20,
                            type = "toggle",
                            name = L["MENU_PER_CLASS_HIGHLIGHTING"],
                            desc = L["MENU_PER_CLASS_HIGHLIGHTING_DESCRIPTION"],
                            get = "IsPerClassHighlighting",
                            set = "TogglePerClassHighlighting",
                        },
                    },
                },
            },
        })

    --
    -- Add our options menu to the Blizzard interface menu.
    --
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addon)
end

function TbChatEnhancements:OnEnable()
    --
    -- Apply all our settings.
    --
    self:ApplyAllSettings()

    --
    -- Condense the combat modifier strings.
    --
    TEXT_MODE_A_STRING_RESULT_ABSORB = L["ABBREVIATION_ABSORB"]
    TEXT_MODE_A_STRING_RESULT_BLOCK = L["ABBREVIATION_BLOCK"]
    TEXT_MODE_A_STRING_RESULT_CRITICAL = L["ABBREVIATION_CRITICAL"]
    TEXT_MODE_A_STRING_RESULT_CRITICAL_SPELL = L["ABBREVIATION_CRITICAL"]
    TEXT_MODE_A_STRING_RESULT_CRUSHING = L["ABBREVIATION_CRUSHING"]
    TEXT_MODE_A_STRING_RESULT_GLANCING = L["ABBREVIATION_GLANCING"]
    TEXT_MODE_A_STRING_RESULT_REFLECT = L["ABBREVIATION_REFLECT"]
    TEXT_MODE_A_STRING_RESULT_RESIST = L["ABBREVIATION_RESIST"]

    --
    -- Register to filter incoming chat events and adjust custom emotes.
    --
    ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", TbChatEnhancements.WhisperInformFilter)
    ChatFrame_AddMessageEventFilter("CHAT_MSG_AFK", TbChatEnhancements.AfkMessageFilter)
    ChatFrame_AddMessageEventFilter("CHAT_MSG_DND", TbChatEnhancements.AfkMessageFilter)
    ChatFrame_AddMessageEventFilter("CHAT_MSG_ACHIEVEMENT", TbChatEnhancements.AchievementMessageFilter)
    ChatFrame_AddMessageEventFilter("CHAT_MSG_GUILD_ACHIEVEMENT", TbChatEnhancements.GuildAchievementMessageFilter)
    ChatFrame_AddMessageEventFilter("CHAT_MSG_SYSTEM", TbChatEnhancements.SystemMessageFilter)
    ChatFrame_AddMessageEventFilter("CHAT_MSG_EMOTE", L["EMOTE_PARSER"])

    --
    -- Reset our stored AFK/DND message if we are neither AFK nor DND on load.
    --
    if (not UnitIsAFK("player") and not UnitIsDND("player")) then self.db.char.afkMessage = "" end

    --
    -- Hook onto the SendChatMessage function so we may toggle our AFK or DND flag before talking.
    --
    self:RawHook("SendChatMessage", "SendChatMessage", true)

    --
    -- Make the officer, raid warning and public channels "sticky".
    --
    ChatTypeInfo["OFFICER"].sticky = 1
    ChatTypeInfo["RAID_WARNING"].sticky = 1
    ChatTypeInfo["CHANNEL"].sticky = 1

    --
    -- Override the chat frames' AddMessage function and hyperlink click handler for chat linking.
    --
    for i = 1, NUM_CHAT_WINDOWS do
        if (_G["ChatFrame" .. i]) then
            self:RawHook(_G["ChatFrame" .. i], "AddMessage", true)
            self:RawHookScript(_G["ChatFrame" .. i], "OnHyperlinkClick")
        end
    end

    --
    -- Reprint the guild message of the day to all channels which display guild chat if we erased it.
    --
    if (not self:IsExtendingChatHistory()) then return end
    local gmotd = GetGuildRosterMOTD()
    if (TbCore:StrIsNotNilOrEmpty(gmotd)) then
        TbCore:PrintForChannel(GUILD_MOTD_LABEL .. " " .. gmotd, "GUILD", "GUILD")
    end
end

function TbChatEnhancements:OnDisable()
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_WHISPER_INFORM", TbChatEnhancements.WhisperInformFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_AFK", TbChatEnhancements.AfkMessageFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_DND", TbChatEnhancements.AfkMessageFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_ACHIEVEMENT", TbChatEnhancements.AchievementMessageFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_GUILD_ACHIEVEMENT", TbChatEnhancements.GuildAchievementMessageFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_SYSTEM", TbChatEnhancements.SystemMessageFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_CHANNEL_NOTICE", TbChatEnhancements.ChannelMessageFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_CHANNEL_NOTICE_USER", TbChatEnhancements.ChannelMessageFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_CHANNEL_JOIN", TbChatEnhancements.ChannelJoinFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_CHANNEL_LEAVE", TbChatEnhancements.ChannelJoinFilter)
    ChatFrame_RemoveMessageEventFilter("CHAT_MSG_EMOTE", L["EMOTE_PARSER"])
end


-------------------------------------------------------------------------------
-- Configure the custom text highlighter.
-------------------------------------------------------------------------------
function TbChatEnhancements:PickHighlightColor()
    --
    -- Setup the color picker frame using the current highlight settings as the default parameters.
    --
    local r, g, b = self:GetHighlightColorRgb()
    ColorPickerFrame.hasOpacity = false
    ColorPickerFrame.previousValues = { r, g, b }
    ColorPickerFrame:SetColorRGB(r, g, b)

    --
    -- Hook into the color picker frame and show it.
    --
    ColorPickerFrame.func = function() self:SetHighlightColorRgb(ColorPickerFrame:GetColorRGB()) end
    ColorPickerFrame.cancelFunc =
        function(previousValues)
            --
            -- Restore our old color values and unhook from the color picker frame.
            --
            ColorPickerFrame.func = nil
            ColorPickerFrame.cancelFunc = nil
            self:SetHighlightColorRgb(unpack(previousValues))
        end
    ColorPickerFrame:Show()
end


-------------------------------------------------------------------------------
-- Properties.
-------------------------------------------------------------------------------
function TbChatEnhancements:GetChatEditBoxLocation()
    return self.db.global.chatEditBoxLocation
end

function TbChatEnhancements:SetChatEditBoxLocation(info, location, ...)
    self.db.global.chatEditBoxLocation = location
    self:ApplyChatEditBoxLocation()
end

function TbChatEnhancements:IsExtendingChatHistory()
    return self.db.global.extendChatHistory
end

function TbChatEnhancements:ToggleExtendChatHistory()
    self.db.global.extendChatHistory = not self:IsExtendingChatHistory()
    self:ApplyExtendChatHistorySetting()
end

function TbChatEnhancements:IsFadingChat()
    return self.db.global.fadeChat
end

function TbChatEnhancements:ToggleFadeChat()
    self.db.global.fadeChat = not self:IsFadingChat()
    self:ApplyFadeChatSetting()
end

function TbChatEnhancements:IsShowingChatFrameButtons()
    return self.db.global.showChatFrameButtons
end

function TbChatEnhancements:ToggleShowChatFrameButtons()
    self.db.global.showChatFrameButtons = not self:IsShowingChatFrameButtons()
    self:ApplyChatFrameButtonsSetting()
end

function TbChatEnhancements:GetChatTimeStampsFormat()
    return self.db.global.chatTimeStampsFormat
end

function TbChatEnhancements:SetChatTimeStampsFormat(info, form, ...)
    self.db.global.chatTimeStampsFormat = form
    self:ApplyChatTimeStampsFormat()
end

function TbChatEnhancements:IsMouseScrolling()
    return self.db.global.mouseScroll
end

function TbChatEnhancements:ToggleMouseScroll()
    self.db.global.mouseScroll = not self:IsMouseScrolling()
    self:ApplyMouseScrollSetting()
end

function TbChatEnhancements:GetHighlightColorRgb()
    return self.db.global.highlightColor["r"], self.db.global.highlightColor["g"], self.db.global.highlightColor["b"]
end

function TbChatEnhancements:SetHighlightColorRgb(r, g, b)
    --
    -- Save our highlight color and cache the hex string representation for it.
    --
    self.db.global.highlightColor["r"] = r
    self.db.global.highlightColor["g"] = g
    self.db.global.highlightColor["b"] = b
    _highlightHex = format("%%s\124c%s%%s\124r%%s", TbCore:RgbToHex(r, g, b))
end

function TbChatEnhancements:GetHighlightWords()
    return TbCore:Capitalize(TbCore:StrJoin(TbCore:GetSortedArray(self.db.profile.highlightWords)))
end

function TbChatEnhancements:SetHighlightWords(info, list, ...)
    self.db.profile.highlightWords = TbCore:StrSplit(TbCore:Capitalize(list), ",")
end

function TbChatEnhancements:IsPerClassHighlighting()
    return self.db.global.perClassHighlighting
end

function TbChatEnhancements:TogglePerClassHighlighting()
    self.db.global.perClassHighlighting = not self:IsPerClassHighlighting()
    self:ApplyPerClassHighlighting()
end

function TbChatEnhancements:IsFilteringChannelAnnouncements()
    return self.db.global.filterChannelAnnouncements
end

function TbChatEnhancements:ToggleFilterChannelAnnouncements()
    self.db.global.filterChannelAnnouncements = not self:IsFilteringChannelAnnouncements()
    self:ApplyChannelAnnouncementFiltering()
end


-------------------------------------------------------------------------------
-- Apply the current settings.
-------------------------------------------------------------------------------
function TbChatEnhancements:ApplyAllSettings()
    self:ApplyChatEditBoxLocation()
    self:ApplyChatFrameButtonsSetting()
    self:ApplyExtendChatHistorySetting()
    self:ApplyFadeChatSetting()
    self:ApplyMouseScrollSetting()
    self:ApplyChatTimeStampsFormat()
    self:ApplyHighlightColor()
    self:ApplyPerClassHighlighting()
    self:ApplyChannelAnnouncementFiltering()
end

function TbChatEnhancements:ApplyChatEditBoxLocation()
    ChatFrameEditBox:ClearAllPoints()
    if (self:GetChatEditBoxLocation() == 1) then
        ChatFrameEditBox:SetPoint("BOTTOMLEFT",  "ChatFrame1", "TOPLEFT",  -5, 0)
        ChatFrameEditBox:SetPoint("BOTTOMRIGHT", "ChatFrame1", "TOPRIGHT", 5, 0)
    else
        ChatFrameEditBox:SetPoint("TOPLEFT",  "ChatFrame1", "BOTTOMLEFT",  -5, 0)
        ChatFrameEditBox:SetPoint("TOPRIGHT", "ChatFrame1", "BOTTOMRIGHT", 5, 0)
    end
end

function TbChatEnhancements:ADDON_LOADED(e, name)
    if (name == "Blizzard_GMChatUI") then
        self:UnregisterEvent("ADDON_LOADED")
        self:ApplyAllSettings()
    end
end

function TbChatEnhancements:ApplyChatFrameButtonsSetting()
    if (self:IsShowingChatFrameButtons()) then
        --
        -- We only want to show the buttons if we were hiding them previously, for compatibility with other chat enhancements add-ons.
        -- If another button is handling them, do not do anything.
        --
        local wasHooked
        for i = 1, NUM_CHAT_WINDOWS do
            if (self:IsHooked(_G["ChatFrame" .. i .. "DownButton"], "OnShow")) then
                wasHooked = true
                self:Unhook(_G["ChatFrame" .. i .. "BottomButton"], "OnShow")
                self:Unhook(_G["ChatFrame" .. i .. "UpButton"], "OnShow")
                self:Unhook(_G["ChatFrame" .. i .. "DownButton"], "OnShow")

                _G["ChatFrame" .. i .. "BottomButton"]:Show()
                _G["ChatFrame" .. i .. "UpButton"]:Show()
                _G["ChatFrame" .. i .. "DownButton"]:Show()
            end
        end
        if (GMChatFrame) then
            if (self:IsHooked(_G["GMChatFrame" .. "DownButton"], "OnShow")) then
                wasHooked = true
                self:Unhook(_G["GMChatFrame" .. "BottomButton"], "OnShow")
                self:Unhook(_G["GMChatFrame" .. "UpButton"], "OnShow")
                self:Unhook(_G["GMChatFrame" .. "DownButton"], "OnShow")

                _G["GMChatFrame" .. "BottomButton"]:Show()
                _G["GMChatFrame" .. "UpButton"]:Show()
                _G["GMChatFrame" .. "DownButton"]:Show()
            end
        end

        if (wasHooked) then
            ChatFrameMenuButton:Show()
        end
    else
        --
        -- We always want to hide the chat frame buttons if the option is unchecked, regardless of other add-ons.
        --
        for i = 1, NUM_CHAT_WINDOWS do
            if (not self:IsHooked(_G["ChatFrame" .. i .. "DownButton"], "OnShow")) then
                self:RawHookScript(_G["ChatFrame" .. i .. "BottomButton"], "OnShow", "HideButton")
                self:RawHookScript(_G["ChatFrame" .. i .. "UpButton"], "OnShow", "HideButton")
                self:RawHookScript(_G["ChatFrame" .. i .. "DownButton"], "OnShow", "HideButton")

                _G["ChatFrame" .. i .. "BottomButton"]:Hide()
                _G["ChatFrame" .. i .. "UpButton"]:Hide()
                _G["ChatFrame" .. i .. "DownButton"]:Hide()
            end
        end
        if (GMChatFrame) then
            if (not self:IsHooked(_G["GMChatFrame" .. "DownButton"], "OnShow")) then
                self:RawHookScript(_G["GMChatFrame" .. "BottomButton"], "OnShow", "HideButton")
                self:RawHookScript(_G["GMChatFrame" .. "UpButton"], "OnShow", "HideButton")
                self:RawHookScript(_G["GMChatFrame" .. "DownButton"], "OnShow", "HideButton")

                _G["GMChatFrame" .. "BottomButton"]:Hide()
                _G["GMChatFrame" .. "UpButton"]:Hide()
                _G["GMChatFrame" .. "DownButton"]:Hide()
            end
        end
        ChatFrameMenuButton:Hide()
    end
end

function TbChatEnhancements:HideButton(frame, ...)
    frame:Hide()
end

function TbChatEnhancements:ApplyExtendChatHistorySetting()
    --
    -- This will clear our chat history.
    --
    if (self:IsExtendingChatHistory()) then
        for num = 1, NUM_CHAT_WINDOWS do
            if (_G["ChatFrame" .. num]) then
                if (_G["ChatFrame" .. num]:GetMaxLines() < 8192) then
                    _G["ChatFrame" .. num]:SetMaxLines(8192)
                end
            end
        end
        if (GMChatFrame and GMChatFrame:GetMaxLines() < 8192) then
            GMChatFrame:SetMaxLines(8192)
        end
    end
end

function TbChatEnhancements:ApplyFadeChatSetting()
    local fade = self:IsFadingChat()
    for num = 1, NUM_CHAT_WINDOWS do
        if (_G["ChatFrame" .. num]) then
            _G["ChatFrame" .. num]:SetFading(fade)
        end
    end
    if (GMChatFrame) then
        GMChatFrame:SetFading(fade)
    end
end

function TbChatEnhancements:ApplyMouseScrollSetting()
    if (self:IsMouseScrolling()) then
        for num = 1, NUM_CHAT_WINDOWS do
            if (_G["ChatFrame" .. num] and not self:IsHooked(_G["ChatFrame" .. num], "OnMouseWheel")) then
                self:RawHookScript(_G["ChatFrame" .. num], "OnMouseWheel")
                _G["ChatFrame" .. num]:EnableMouseWheel(true)
            end
        end
    else
        for num = 1, NUM_CHAT_WINDOWS do
            if (_G["ChatFrame" .. num] and self:IsHooked(_G["ChatFrame" .. num], "OnMouseWheel")) then
                self:Unhook(_G["ChatFrame" .. num], "OnMouseWheel")
            end
        end
    end
end

function TbChatEnhancements:ApplyChatTimeStampsFormat()
    local form = self:GetChatTimeStampsFormat()
    _chatTimeStampsFormat = "%H:%M:%S>"
    if (form == 5) then
        _chatTimeStampsFormat = "[%I:%M:%S %p]"
    elseif (form == 10) then
        _chatTimeStampsFormat = "%I:%M:%S %p>"
    elseif (form == 15) then
        _chatTimeStampsFormat = "[%H:%M:%S]"
    end
end

function TbChatEnhancements:ApplyHighlightColor()
    self:SetHighlightColorRgb(self:GetHighlightColorRgb())
end

function TbChatEnhancements:ApplyPerClassHighlighting()
    local operation = self.UnregisterEvent
    if (self:IsPerClassHighlighting()) then
        operation = self.RegisterEvent
    end

    --
    -- We do not register for PLAYER_GUILD_UPDATE to request an updated guild roster, because TB Core takes care of that for us.
    --
    operation(self, "CHAT_MSG_SYSTEM")
    operation(self, "UPDATE_MOUSEOVER_UNIT")
    operation(self, "PLAYER_TARGET_CHANGED")
    operation(self, "WHO_LIST_UPDATE")
    operation(self, "PARTY_MEMBERS_CHANGED")
    operation(self, "UPDATE_BATTLEFIELD_STATUS")
    operation(self, "GUILD_ROSTER_UPDATE")
    operation(self, "FRIENDLIST_UPDATE")
    operation(self, "ZONE_CHANGED_NEW_AREA")
    operation(self, "CHAT_MSG_CHANNEL")
    operation(self, "CHAT_MSG_WHISPER", "CHAT_MSG_CHANNEL")
end

function TbChatEnhancements:ApplyChannelAnnouncementFiltering()
    local operation = ChatFrame_RemoveMessageEventFilter
    if (self:IsFilteringChannelAnnouncements()) then operation = ChatFrame_AddMessageEventFilter end

    operation("CHAT_MSG_CHANNEL_NOTICE", TbChatEnhancements.ChannelMessageFilter)
    operation("CHAT_MSG_CHANNEL_NOTICE_USER", TbChatEnhancements.ChannelMessageFilter)
    operation("CHAT_MSG_CHANNEL_JOIN", TbChatEnhancements.ChannelJoinFilter)
    operation("CHAT_MSG_CHANNEL_LEAVE", TbChatEnhancements.ChannelJoinFilter)
end


-------------------------------------------------------------------------------
-- Miscellaneous helper function.
-------------------------------------------------------------------------------
function TbChatEnhancements:CleanUpPlayerCache()
    local t = time()
    for name, data in pairs(self.db.realm.players) do
        if (type(data) ~= "table") then
            --
            -- Upgrade entries from previous versions of TB Chat Enhancements to include a time stamp of when they were added.
            --
            self.db.realm.players[name] = { ["class"] = data, ["time"] = t }
        elseif (t - data["time"] > 15552000 or strfind(name, "%-$")) then
            --
            -- Delete entries older than 60 seconds * 60 minutes * 24 hours * 30 days * 6 months = 15552000 seconds, or which end
            -- in "-". The latter resulted from rare cases where the realm name of a player from our realm was an empty string rather
            -- than nil.
            --
            self.db.realm.players[name] = nil
        end
    end
end

--
-- Get the chat timestamp hyperlink prefix.
--
function TbChatEnhancements:GetCopyChatPrefix()
    return "TbCCP:"
end

function TbChatEnhancements:GetCopyChatItemPrefix()
    return "TbCCI:"
end

function TbChatEnhancements:GetCopyChatAchievementPrefix()
    return "TbCCA:"
end

function TbChatEnhancements:GetCopyChatSpellPrefix()
    return "TbCCS:"
end

function TbChatEnhancements:IsCrossGuildChatWhispering()
    return TbCrossGuildChat and TbCrossGuildChat:IsEmitting()
end

function TbChatEnhancements:IsAutoBGWhispering()
    if (TbAutoBG) then
        return TbCore:StrIsNotNilOrEmpty(TbAutoBG:GetIdleNotifyTarget())
    end
end
