﻿--[[
    Author: Valana <Team BATTLE> of Mug'thol Alliance-US
]]

--
-- Must never be local: relied upon by TB Add-Ons!
--
TbCore = LibStub("AceAddon-3.0"):NewAddon("TbCore", "AceConsole-3.0", "AceComm-3.0", "AceEvent-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("TB_Core", true)
local Z = LibStub("LibBabble-Zone-3.0"):GetLookupTable()

local _G = _G
local MAX_BATTLEFIELD_QUEUES = MAX_BATTLEFIELD_QUEUES
local NUM_CHAT_WINDOWS = NUM_CHAT_WINDOWS
local UIERRORS_HOLD_TIME = UIERRORS_HOLD_TIME
local RAID_CLASS_COLORS = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
local tostring = tostring
local type = type
local pairs = pairs
local tinsert = tinsert
local sort = sort
local strupper = strupper
local strlower = strlower
local strfind = strfind
local strlen = strlen
local strsub = strsub
local gsub = gsub
local strmatch = strmatch
local format = format
local abs = abs
local floor = floor
local ceil = ceil
local min = min
local max = max
local wipe = wipe
local GetChatWindowMessages = GetChatWindowMessages
local UIErrorsFrame = UIErrorsFrame
local RaidNotice_AddMessage = RaidNotice_AddMessage
local RaidBossEmoteFrame = RaidBossEmoteFrame
local ChatTypeInfo = ChatTypeInfo
local PlaySound = PlaySound
local IsRaidLeader = IsRaidLeader
local IsRaidOfficer = IsRaidOfficer
local UnitExists = UnitExists
local UnitIsUnit = UnitIsUnit
local UnitInRaid = UnitInRaid
local UnitIsPartyLeader = UnitIsPartyLeader
local GetNumPartyMembers = GetNumPartyMembers
local GetNumRaidMembers = GetNumRaidMembers
local GetRealZoneText = GetRealZoneText
local GetSubZoneText = GetSubZoneText
local GetWorldStateUIInfo = GetWorldStateUIInfo
local IsInInstance = IsInInstance
local GetBattlefieldStatus = GetBattlefieldStatus
local GetGuildInfo = GetGuildInfo
local GetNumGuildMembers = GetNumGuildMembers
local SetGuildRosterShowOffline = SetGuildRosterShowOffline
local GetGuildRosterInfo = GetGuildRosterInfo
local GetNumFriends = GetNumFriends
local GetFriendInfo = GetFriendInfo
local GetNumIgnores = GetNumIgnores
local GetIgnoreName = GetIgnoreName
local GetSpellCooldown = GetSpellCooldown
local GetCompanionInfo = GetCompanionInfo
local GetNumCompanions = GetNumCompanions

local _player = UnitName("player")
local _isValana = GetRealmName() == "Mug'thol" and (_player == "Valana" or _player == "Givenchy" or _player == "Karuna" or _player == "Iol")
local _isProcessingGuildRosterUpdate
local _capitalizations
local _caseInsensitivePatterns
local _numGuildMates
local _guildMates = { }
local _numFriends
local _friends = { }
local _numIgnores
local _ignores = { }
local _colors =
{
    RED     = { r = 1, g = 0, b = 0 }, ORANGE = { r = 1, g = .5, b = 0 },
    YELLOW  = { r = 1, g = 1, b = 0 }, GREEN  = { r = 0, g = 1, b = 0 },
    LTBLUE  = { r = 0, g = 1, b = 1 }, BLUE   = { r = 0, g = 0, b = 1 },
    PURPLE  = { r = 1, g = 0, b = 1 }, WHITE  = { r = 1, g = 1, b = 1 },
    BLACK   = { r = 0, g = 0, b = 0 }, PINK   = { r = 1, g = .5, b = 1 },
    GUILD   = ChatTypeInfo["GUILD"],
    OFFICER = ChatTypeInfo["OFFICER"],
    SYSTEM = ChatTypeInfo["SYSTEM"],
    WHISPER = ChatTypeInfo["WHISPER"],
    DEATHKNIGHT = RAID_CLASS_COLORS["DEATHKNIGHT"],
    DRUID = RAID_CLASS_COLORS["DRUID"],
    HUNTER = RAID_CLASS_COLORS["HUNTER"],
    MAGE = RAID_CLASS_COLORS["MAGE"],
    PALADIN = RAID_CLASS_COLORS["PALADIN"],
    PRIEST = RAID_CLASS_COLORS["PRIEST"],
    ROGUE = RAID_CLASS_COLORS["ROGUE"],
    SHAMAN = RAID_CLASS_COLORS["SHAMAN"],
    WARLOCK = RAID_CLASS_COLORS["WARLOCK"],
    WARRIOR = RAID_CLASS_COLORS["WARRIOR"],
}


-------------------------------------------------------------------------------
-- Initialization.
-------------------------------------------------------------------------------
function TbCore:OnInitialize()
    --
    -- Define shorter aliases for the functions provided in this library.
    --
    function TPrint(msg, frameNumber, color) return self:Print(msg, frameNumber, color) end
    function TDebug(msg, frameNumber, color) return self:Debug(msg, frameNumber, color) end
    function TEPrint(msg, color, t) return self:PrintError(msg, color, t) end

    --
    -- Initialize our metatables
    --
    _capitalizations = setmetatable({ }, { __index =
        function(tbl, phrase)
            local capitalized = self:StrTrim(gsub(phrase, "%f[%a\128-\255%*]([%a\128-\255%*])([%a\128-\255%*]*)%f[^%a\128-\255%*]", function(firstLetter, word) return strupper(firstLetter) .. strlower(word) end))
            rawset(tbl, phrase, capitalized)
            return capitalized
        end })

    _caseInsensitivePatterns = setmetatable ({ }, { __index =
        function(tbl, phrase)
            local pattern = gsub(phrase, "[%a\128-\255]",
                function(ch)
                    local lower, upper = strlower(ch), strupper(ch)
                    if (lower == upper) then return ch end
                    return "[" .. lower .. upper .. "]"
                end)
            rawset(tbl, phrase, pattern)
            return pattern
        end })

    --
    -- Define the /gx commands.
    --
    self:RegisterChatCommand(L["CMD_RESTART_GX1"], RestartGx)
    self:RegisterChatCommand(L["CMD_RESTART_GX2"], RestartGx)
    self:RegisterChatCommand(L["CMD_RESTART_GX3"], RestartGx)

    --
    -- Define the /rl commands.
    --
    self:RegisterChatCommand(L["CMD_RELOAD_UI1"], ReloadUI)
    self:RegisterChatCommand(L["CMD_RELOAD_UI2"], ReloadUI)
end

function TbCore:OnEnable()
--[===[@alpha@
    if (not _isValana) then
        --
        -- Re-enable essential addons disabled by the dumbasses.
        --
        if (IsAddOnLoaded("FuBar")) then
            EnableAddOn("Broker2Fubar")
            EnableAddOn("TB_AutoBG")
        end
        EnableAddOn("TB_AltTrader")
        EnableAddOn("TB_Artisan")
        EnableAddOn("TB_ChatEnhancements")
        EnableAddOn("TB_Misdirector")
        EnableAddOn("TB_RaidHelper")

        --
        -- Toggle class-specific addons.
        --
        local level, _, class = UnitLevel("player"), UnitClass("player")
        if (class == "HUNTER") then
            if (level > 19) then EnableAddOn("TB_AntiDaze") end
            if (level > 9) then EnableAddOn("TB_AutoTracker") end
        elseif (class == "WARRIOR" and IsUsableSpell(GetSpellInfo(50720))) then
            EnableAddOn("TB_Vigilante")
        elseif ((class == "DRUID" and level > 19) or ((class == "PALADIN" or class == "SHAMAN") and level > 29)) then
            EnableAddOn("TB_HeroicLust")
        end
    end

    --
    -- Hide the action bar gryphons and the minimap world map button.
    --
    MainMenuBarLeftEndCap:Hide()
    MainMenuBarRightEndCap:Hide()
    MiniMapWorldMapButton:Hide()
--@end-alpha@]===]

    --
    -- Make World of WarCraft use all cores if more than 2 are detected.
    --
    local cores = tonumber(GetCVar("coresDetected"))
    if (cores > 2) then
        SetCVar("processAffinityMask", math.pow(cores, 2) - 1)
    end

    --
    -- Increase the camera's maximum follow distance.
    --
    SetCVar("CameraDistanceMax", 50)        -- default: 15
    SetCVar("CameraDistanceMoveSpeed", 100) -- default: 8.33

    --
    -- Always disable vertical synchronization.
    --
    SetCVar("GxVSync", 0)

    --
    -- Show enemy cast bars.
    --
    SetCVar("ShowTargetCastbar", 1)
    SetCVar("ShowVKeyCastbar", 1)

    --
    -- Increase the quality of screenshots to 80% for JPG's (default: 30%).
    --
    if (tonumber(GetCVar("ScreenshotQuality")) < 8) then SetCVar("ScreenshotQuality", 8) end

    --
    -- Register to cache the guild roster and friends list, and request the guild roster.
    --
    self:RegisterEvent("GUILD_ROSTER_UPDATE")
    self:RegisterEvent("PLAYER_GUILD_UPDATE")
    self:RegisterEvent("FRIENDLIST_UPDATE")
    self:RegisterEvent("IGNORELIST_UPDATE")

    --
    -- Improve the IsFlyableArea and CallCompanion functions.
    --
    self:RawHook("CallCompanion", "CallCompanion", true)
    self:RawHook("IsFlyableArea", "IsFlyableArea", true)

    --
    -- Register for communications received from AutoBG.
    --
    self:RegisterComm("AUTOBG_NOTIFY_GM", "AUTOBG_NOTIFY_GM")
    self:RegisterComm("AUTOBG_NOTIFY_IDLE", "AUTOBG_NOTIFY_IDLE")
    self:RegisterComm("AUTOBG_NOTIFY_INACTIVE", "AUTOBG_NOTIFY_INACTIVE")
end


-------------------------------------------------------------------------------
-- Printers.
-------------------------------------------------------------------------------
--
-- Print a message to a chat window.
--
function TbCore:Print(msg, frameNumber, color, hide)
    if (not msg or hide) then return end
    msg = tostring(msg)

    --
    -- Get the RGB components of the requested display color for the message.
    --
    local red, green, blue = self:ColorToRgb(color)

    --
    -- Determine which frame to send the message to.
    --
    if (not frameNumber or type(frameNumber) ~= "number" or frameNumber > NUM_CHAT_WINDOWS) then
        frameNumber = 1
    end
    _G["ChatFrame" .. frameNumber]:AddMessage(msg, red or 1, green or 1, blue or 1)
end

--
-- Print a debug message to a chat window.
--
function TbCore:Debug(msg, frameNumber, color)
    self:Print(msg, frameNumber, color, _isValana)
end

--
-- Print to every chat frame displaying the specified channel.
--
function TbCore:PrintForChannel(message, channel, color)
    for i = 1, NUM_CHAT_WINDOWS do
        self:PrintForChannelHelper(message, channel, i, color, GetChatWindowMessages(i))
    end
end

function TbCore:PrintForChannelHelper(message, channel, index, color, accepted, ...)
    if (accepted == channel) then
        self:Print(message, index, color)
    elseif (accepted) then
        self:PrintForChannelHelper(message, channel, index, color, ...)
    end
end

--
-- Print a message to the error frame.
--
function TbCore:PrintError(msg, color, holdTime)
    if (not msg) then return end
    msg = tostring(msg)

    --
    -- Get the hex triplet of the requested display color for the message.
    --
    local red, green, blue = self:ColorToRgb(color)

    --
    -- Display the message in the error frame using the requested color and display time,
    -- or defaults if unspecified.
    --
    UIErrorsFrame:AddMessage(msg, red or 1, green or 1, blue or 1, holdTime or UIERRORS_HOLD_TIME)
end

--
-- Print a message to the raid boss emote frame.
--
function TbCore:PrintRaidBossEmote(message, playSound)
    RaidNotice_AddMessage(RaidBossEmoteFrame, message, ChatTypeInfo["RAID_BOSS_EMOTE"])
    if (playSound) then PlaySound("RaidWarning") end
end


-------------------------------------------------------------------------------
-- Group utility methods.
-------------------------------------------------------------------------------
--
-- Figure out if we have permission to invite people.
--
function TbCore:CanInvite()
    return not TbCore:IsInGroup() or ((IsRaidLeader() or IsRaidOfficer()) and self:GroupHasRoom())
end

--
-- Figure out if we are able to use the Raid Warning channel.
--
function TbCore:CanUseRaidWarning()
    return TbCore:IsInParty() or (TbCore:IsInRaid("player") and (IsRaidLeader() or IsRaidOfficer()))
end

--
-- Figure out if we can uninvite the specified unit.
--
function TbCore:CanUninviteUnit(unit)
    return UnitExists(unit) and not UnitIsUnit("player", unit) and self:IsUnitInGroup(unit) and (IsRaidLeader() or (IsRaidOfficer() and not UnitIsPartyLeader(unit)))
end

--
-- Figure out if we have room for more players.
--
function TbCore:GroupHasRoom()
    return (not TbCore:IsInGroup() or (not TbCore:IsInRaid() and GetNumPartyMembers() < 4) or GetNumRaidMembers() < 40)
end

--
-- Get the string representation of our group type.
--
function TbCore:GetRaidOrParty()
    if (TbCore:IsInRaid()) then return "RAID" end
    if (GetNumPartyMembers() > 0) then return "PARTY" end
end

--
-- Determines whether we are in a group.
--
function TbCore:IsInGroup()
    return TbCore:IsInRaid() or GetNumPartyMembers() > 0
end

--
-- Determines whether we are in a 5-player party.
--
function TbCore:IsInParty()
    return not TbCore:IsInRaid() and GetNumPartyMembers() > 0
end

--
-- Determines whether we are in a raid.
--
function TbCore:IsInRaid()
    return GetNumRaidMembers() > 0
end

--
-- Figure out if a unit is in our group.
--
function TbCore:IsUnitInGroup(unit)
    return UnitInRaid(unit) or UnitInParty(unit)
end


-------------------------------------------------------------------------------
-- Arena / battleground / instance helper functions.
-------------------------------------------------------------------------------
--
-- Figure out if we are currently in Lake Wintergrasp.
--
function TbCore:IsInWintergrasp()
    return GetRealZoneText() == Z["Wintergrasp"]
end

--
-- Figure out if we are currently in an ongoing Lake Wintergrasp battle.
--
function TbCore:IsInActiveWintergrasp()
    if (self:IsInWintergrasp() and not GetWintergraspWaitTime()) then return true end
end

--
-- Figure out if we are currently in a battleground.
--
function TbCore:IsInBattleground()
    local _, t = IsInInstance()
    return t == "pvp"
end

--
-- Figure out if we are currently in an arena.
--
function TbCore:IsInArena()
    local _, t = IsInInstance()
    return t == "arena"
end

--
-- Figure out if we are currently waiting in a rated arena queue.
--
function TbCore:IsInRatedArenaQueue()
    for i = 1, MAX_BATTLEFIELD_QUEUES do
        local status, _, _, _, _, _, rated = GetBattlefieldStatus(i)
        if (rated == 1 and status == "confirm") then return true end
    end
end

--
-- Figure out if we are currently in a PvP instance.
--
function TbCore:IsInPvPInstance()
    return self:IsInBattleground() or self:IsInArena()
end

--
-- Figure out if we are currently in a 5-person dungeon.
--
function TbCore:IsInDungeon()
    local _, t = IsInInstance()
    return t == "party"
end

--
-- Figure out if we are currently in a raid dungeon.
--
function TbCore:IsInRaidInstance()
    local _, t = IsInInstance()
    return t == "raid"
end

--
-- Figure out if we are currently in a PvE instance.
--
function TbCore:IsInPvEInstance()
    return self:IsInDungeon() or self:IsInRaidInstance()
end

--
-- Figure out if we are currently in an instance.
--
function TbCore:IsInInstance()
    return not self:IsNotInInstance()
end

function TbCore:IsNotInInstance()
    local _, t = IsInInstance()
    return t == "none"
end


-------------------------------------------------------------------------------
-- Mount Methods.
-------------------------------------------------------------------------------
--
-- Determine if we truly are in a flyable area.
--
function TbCore:IsFlyableArea()
    local isFlyableArea = self.hooks["IsFlyableArea"]()
    if (not isFlyableArea) then return isFlyableArea end
    return IsFlying() or (not self:IsInWintergrasp() and (GetRealZoneText() ~= Z["Dalaran"] or GetSubZoneText() == Z["Krasus' Landing"]))
end

--
-- Call a companion either by index or by name.
--
function TbCore:CallCompanion(kind, index)
    --
    -- If this is a regular call, pass it on to the Blizzard function.
    --
    if (type(index) == "number") then
        self.hooks["CallCompanion"](kind, index)
        return true
    end

    --
    -- Look for the companion of the specified kind and with the specified name.
    --
    if ((kind == "MOUNT" or kind == "CRITTER") and type(index) == "string") then
        for i = 1, GetNumCompanions(kind) do
            local _, name = GetCompanionInfo(kind, i)
            if (name == index) then
                self.hooks["CallCompanion"](kind, i)
                return true
            end
        end
        return
    end

    --
    -- Try finding a mount with the specified name, then a critter if there is no such mount.
    --
    if (self:CallCompanion("MOUNT", kind)) then return true end
    return self:CallCompanion("CRITTER", kind)
end


-------------------------------------------------------------------------------
-- Friends and Guild Methods.
-------------------------------------------------------------------------------
function TbCore:PLAYER_GUILD_UPDATE(e)
    if (IsInGuild()) then
        if (_numGuildMates == 0) then
            GuildRoster()
        end
    else
        _numGuildMates = 0
        wipe(_guildMates)
    end
end

function TbCore:GUILD_ROSTER_UPDATE(e, isLocal)

    --
    -- When a player joins or leaves the guild, 1 is passed as the value of isLocal and we do not automatically get the updated
    -- guild roster, so we should request it. Other TB addons should not call GuildRoster themselves to avoid unnecessary traffic.
    --
    if (isLocal) then
        GuildRoster()
    elseif (not _isProcessingGuildRosterUpdate) then
        --
        -- Changing whether we are showing offline members will cause this event to re-enter with isLocal being nil, so check our flag.
        --
        _isProcessingGuildRosterUpdate = true
        self:CheckGuildMatesPopulated(true)
        _isProcessingGuildRosterUpdate = nil
    end
end

function TbCore:CheckGuildMatesPopulated(populate)
    if (_numGuildMates == 0) then
        populate = true
    end
    if (populate) then
        --
        -- Rebuild our cache if we have a different number of guild mates compared to the previous check.
        --
        local numGuildMates = GetNumGuildMembers(true)
        if (numGuildMates == _numGuildMates) then return end

        local wasShowingOffling = GetGuildRosterShowOffline()
        SetGuildRosterShowOffline(true)
        wipe(_guildMates)
        for i = 1, numGuildMates do
            local name = GetGuildRosterInfo(i)
            if (not name) then
                --
                -- If the guild mate's data was unavailable, exit without updating our guild mate count. This was most likely because roster
                -- information was not yet fully available.
                --
                SetGuildRosterShowOffline(wasShowingOffline)
                _isProcessingGuildRosterUpdate = nil
                return
            end
            _guildMates[name] = true
        end
        SetGuildRosterShowOffline(wasShowingOffline)
        _numGuildMates = numGuildMates
    end
end

function TbCore:FRIENDLIST_UPDATE(e)
    self:CheckFriendsPopulated(true)
end

function TbCore:CheckFriendsPopulated(populate)
    if (not _numFriends) then
        populate = true
    end
    if (populate) then
        --
        -- Rebuild our cache if we have a different number of friends compared to the previous check.
        --
        local numFriends = GetNumFriends()
        if (numFriends == _numFriends) then return end
        _numFriends = numFriends

        wipe(_friends)
        for i = 1, _numFriends do
            _friends[GetFriendInfo(i)] = true
        end
    end
end

function TbCore:IGNORELIST_UPDATE(e)
    self:CheckIgnoresPopulated(true)
end

function TbCore:CheckIgnoresPopulated(populate)
    if (not _numIgnores) then
        populate = true
    end
    if (populate) then
        --
        -- Rebuild our cache if we have a different number of ignored players compared to the previous check.
        --
        local numIgnores = GetNumIgnores()
        if (numIgnores == _numIgnores) then return end
        _numIgnores = numIgnores

        wipe(_ignores)
        for i = 1, _numIgnores do
            _ignores[GetIgnoreName(i)] = true
        end
    end
end

function TbCore:IsFriendOrGuildMate(unit)
    unit = self:Capitalize(UnitName(unit or "") or unit or "")
    if (not unit) then return end

    self:CheckFriendsPopulated()
    if (_friends[unit]) then return true end

    if (UnitIsUnit("player", unit)) then return end
    self:CheckGuildMatesPopulated()
    return _guildMates[unit]
end

function TbCore:IsGuildMate(unit)
    if (not IsInGuild()) then return end
    unit = unit or ""
    unit = self:Capitalize(UnitName(unit) or unit)
    if (not unit or UnitIsUnit("player", unit)) then return end

    self:CheckGuildMatesPopulated()
    return _guildMates[unit]
end

function TbCore:IsFriend(unit)
    unit = unit or ""
    unit = self:Capitalize(UnitName(unit) or unit)
    if (not unit) then return end

    self:CheckFriendsPopulated()
    return _friends[unit]
end

function TbCore:IsIgnored(unit)
    unit = unit or ""
    unit = self:Capitalize(UnitName(unit) or unit)
    if (not unit) then return end

    --
    -- Check the Silent Ignore list if applicable, then the default ignore list.
    --
    if (TbSilentIgnore and TbSilentIgnore:IsIgnored(unit)) then return true end
    self:CheckIgnoresPopulated()
    return _ignores[unit]
end

--
-- Get the unlocalized string representation of the specified localized class name.
--
function TbCore:GetUnlocalizedClass(localized)
    return L[localized]
end


-------------------------------------------------------------------------------
-- Color Utility Methods.
-------------------------------------------------------------------------------
--
-- Convert the color with the specified name to its RGB components.
--
function TbCore:ColorToRgb(color)
    --
    -- Default to white.
    --
    local c = _colors[strupper(color or "WHITE")]
    if (c) then return c.r, c.g, c.b end
end

--
-- Convert a color's RGB components to its hexadecimal string representation.
--
function TbCore:RgbToHex(r, g, b, a)
    return format("%02x%02x%02x%02x", self:ClampRgbValue(a, 1) * 255, self:ClampRgbValue(r, 0) * 255, self:ClampRgbValue(g, 0) * 255, self:ClampRgbValue(b, 0) * 255)
end

function TbCore:ClampRgbValue(component, default)
    component = component or default or 0
    return max(0, min(component, 255))
end


-------------------------------------------------------------------------------
-- Spell Utility Methods.
-------------------------------------------------------------------------------
--
-- Find the remaining amount of time remaining on the specified spell, in seconds.
--
function TbCore:GetRemainingCooldown(spell)
    if (type(spell) == "number") then
        spell = GetSpellInfo(spell)
    end
    local startTime, cooldown = GetSpellCooldown(spell)
    if (not startTime) then return end
    return max(0, startTime + cooldown - GetTime())
end

--
-- Determine if we have a certain talent.
--
function TbCore:HasTalent(tree, slot)
    local _, _, _, _, rank = GetTalentInfo(tree, slot)
    return (rank > 0)
end


-------------------------------------------------------------------------------
-- Translation String Utility Methods.
-------------------------------------------------------------------------------
function TbCore:GetHeOrShe(unit)
    if (UnitSex(unit) == 3) then return L["SHE"] end
    return L["HE"]
end

--
-- Determine the appropriate possessive string.
--
function TbCore:GetHisOrHer(unit)
    if (UnitSex(unit) == 3) then return L["HER"] end
    return L["HIS"]
end

--
-- Determine the appropriate pronoun string.
--
function TbCore:GetHimOrHer(unit)
    if (UnitSex(unit) == 3) then return L["HER"] end
    return L["HIM"]
end

--
-- Determine the appropriate reflective string.
--
function TbCore:GetHimselfOrHerself(unit)
    if (UnitSex(unit) == 3) then return L["HERSELF"] end
    return L["HIMSELF"]
end


-------------------------------------------------------------------------------
-- Table and String Utility Methods.
-------------------------------------------------------------------------------
--
-- Count the number of entries in the specified dictionary.
--
function TbCore:CountDictionaryEntries(dictionary)
    --
    -- First check if this is an array. Note that if it is a mixed table, the count will be incorrect.
    --
    local count = #dictionary
    if (count > 0) then return count end

    --
    -- Do an actual count of the number of entries.
    --
    for key, value in pairs(dictionary) do count = count + 1 end
    return count
end

--
-- Build a sorted array from the specified dictionary.
--
function TbCore:GetSortedArray(dictionary, comparer)
    local array = dictionary
    if (#dictionary == 0) then
        array = { }
        for key, value in pairs(dictionary) do
            tinsert(array, key)
        end
    end

    sort(array, comparer)
    return array
end

--
-- Determine whether a string is nil or entirely made up of white space.
--
function TbCore:StrIsNilOrEmpty(str)
    if (not str or not strfind(str, "[^%s]")) then return true end
end

function TbCore:StrIsNotNilOrEmpty(str)
    if (str and strfind(str, "[^%s]")) then return true end
end

--
-- Trim white spaces and the specified delete string from the start and end of the specified string if it's not nil.
--
function TbCore:StrTrim(value, delete)
    if (not value) then return end

    if (self:StrIsNotNilOrEmpty(delete)) then
        delete = "%s*" .. delete .. "*%s*"
    else
        delete = "%s*"
    end

    --
    -- Remove the pattern from the start and from the end, then trim the result.
    --
    return gsub(value, "^" .. delete .. "(.-)" .. delete .. "$", "%1")
end

function TbCore:StrLowerTrim(value, delete)
    if (not value) then return end
    return strlower(self:StrTrim(value, delete))
end

--
-- Capitalize the first letter of every word in the specified string, while converting all other letters to lower case.
--
function TbCore:Capitalize(words)
    if (not words) then return end

    return _capitalizations[words]
end

--
-- Get a pattern which can be used to search for the specified word case-insensitively.
--
function TbCore:GetCaseInsensitivePattern(phrase)
    if (not phrase) then return end

    return _caseInsensitivePatterns[phrase]
end

--
-- Extract the visible text from a hyperlink, or return the parameter itself if it is not a hyperlink.
--
function TbCore:GetTextFromHyperlink(link)
    if (not link) then return end
    link = tostring(link)

    --
    -- Parse real hyperlinks.
    --
    local text = strmatch(link, "\124H.-\124h%[(.+)%]\124h")
    if (not text) then
        --
        -- Try to parse fake hyperlinks by removing any brackets.
        --
        text = strmatch(link, "%[*([^]]+)")
    end
    return self:StrTrim(text)
end

--
-- Split the specified string into a table of words around the specified delimiter.
--
function TbCore:StrSplit(value, delimiter, createDictionary)
    --
    -- Build a table and look for the first delimiter.
    --
    delimiter = delimiter or " "
    value = self:StrTrim(value)
    local split = { }
    local add = function(v) tinsert(split, v) end
    if (createDictionary) then
        add = function(v) split[v] = true end
    end
    local index = strfind(value, delimiter)
    while (index) do
        --
        -- Add the first token to our array and remove it from our string.
        --
        if (index > 1) then
            add(strsub(value, 1, index - 1))
        end
        value = self:StrTrim(strsub(value, index + 1))
        index = strfind(value, delimiter)
    end

    --
    -- Add the last token to our table.
    --
    if (TbCore:StrIsNotNilOrEmpty(value)) then add(value) end
    return split
end

--
-- Join the elements from the specified array into a string delimited by the specified delimiter, returning the specified value when the array is empty.
--
function TbCore:StrJoin(array, delimiter, whenEmpty)
    delimiter = delimiter or ", "
    local str = ""
    for k, v in pairs(array) do
        --
        -- The table might be a dictionary or an array, or a mixed-indexed table.
        --
        local element = ""
        if (type(k) == "string") then
            element = self:StrTrim(k)
        elseif (type(v) == "string") then
            element = self:StrTrim(v)
        end
        if (self:StrIsNotNilOrEmpty(element)) then str = str .. delimiter .. element end
    end
    if (self:StrIsNilOrEmpty(str)) then
        str = whenEmpty or ""
    else
        str = strsub(str, strlen(delimiter) + 1)
    end
    return str
end

--
-- Converts a time specified in seconds into a more legible format.
--
function TbCore:GetFormattedTime(value)
    --
    -- Calculate the number of hours.
    --
    local hours, minutes
    local seconds = abs(value)
    hours = floor(seconds / 3600)
    seconds = seconds % 3600

    --
    -- Calculate the number of minutes. What remains is the number of seconds.
    --
    minutes = floor(seconds / 60)
    seconds = ceil(seconds % 60)

    local formatted = ""
    if (value < 0) then formatted = "-" end
    if (hours > 0) then formatted = formatted .. hours .. L["H"] .. " " end
    if (minutes > 0) then formatted = formatted .. minutes .. L["MN"] .. " " end
    if (seconds > 0) then formatted = formatted .. seconds .. L["S"] end

    return self:StrTrim(formatted)
end


-------------------------------------------------------------------------------
-- Event handlers for communications received from AutoBG.
-------------------------------------------------------------------------------
function TbCore:AUTOBG_NOTIFY_GM(prefix, message, distribution, sender)
    --
    -- Only react to notifications from friends and guildmates as a safeguard.
    --
    if (self:IsIgnored(sender) or not self:IsFriendOrGuildMate(sender)) then return end

    self:PrintAutoBGAlert(L["AUTOBG_GM_NOTIFY"])
end

function TbCore:AUTOBG_NOTIFY_IDLE(prefix, message, distribution, sender)
    --
    -- Only react to notifications from friends and guildmates as a safeguard.
    --
    if (self:IsIgnored(sender) or not self:IsFriendOrGuildMate(sender)) then return end

    self:PrintAutoBGAlert(format(L["AUTOBG_IDLE_NOTIFY"], sender))
end

function TbCore:AUTOBG_NOTIFY_INACTIVE(prefix, message, distribution, sender)
    --
    -- Only react to notifications from friends and guildmates as a safeguard.
    --
    if (self:IsIgnored(sender) or not self:IsFriendOrGuildMate(sender)) then return end

    self:PrintAutoBGAlert(format(L["AUTOBG_INACTIVE_NOTIFY"], sender))
end

function TbCore:PrintAutoBGAlert(message)
    --
    -- Print it to all chat windows in pink.
    --
    for i = 1, NUM_CHAT_WINDOWS do
        self:Print(message, i, "WHISPER")
    end

    --
    -- Only play the alert sound if not in an arena to avoid distracting the player.
    --
    if (not self:IsInArena()) then
        self:PrintRaidBossEmote(message, true)
    end
end
